KEY_0 = 0x30
KEY_1 = 0x31
KEY_2 = 0x32
KEY_3 = 0x33
KEY_4 = 0x34
KEY_5 = 0x35
KEY_6 = 0x36
KEY_7 = 0x37
KEY_8 = 0x38
KEY_9 = 0x39
KEY_A = 0x41
KEY_B = 0x42
KEY_C = 0x43
KEY_D = 0x44
KEY_E = 0x45
KEY_F = 0x46
KEY_G = 0x47
KEY_H = 0x48
KEY_I = 0x49
KEY_J = 0x4A
KEY_K = 0x4B
KEY_L = 0x4C
KEY_M = 0x4D
KEY_N = 0x4E
KEY_O = 0x4F
KEY_P = 0x50
KEY_Q = 0x51
KEY_R = 0x52
KEY_S = 0x53
KEY_T = 0x54
KEY_U = 0x55
KEY_V = 0x56
KEY_W = 0x57
KEY_X = 0x58
KEY_Y = 0x59
KEY_Z = 0x5A

VK_LBUTTON = 0x01               # Left mouse button
VK_RBUTTON = 0x02               # Right mouse button
VK_CANCEL = 0x03                # Control-break processing
VK_MBUTTON = 0x04               # Middle mouse button (three-button mouse)
VK_XBUTTON1 = 0x05              # X1 mouse button
VK_XBUTTON2 = 0x06              # X2 mouse button
VK_BACK = 0x08                  # BACKSPACE key
VK_TAB = 0x09                   # TAB key
VK_CLEAR = 0x0C                 # CLEAR key
VK_RETURN = 0x0D                # ENTER key
VK_ENTER = VK_RETURN
VK_SHIFT = 0x10                 # SHIFT key
VK_CONTROL = 0x11               # CTRL key
VK_MENU = 0x12                  # ALT key
VK_PAUSE = 0x13                 # PAUSE key
VK_CAPITAL = 0x14               # CAPS LOCK key
VK_KANA = 0x15                  # IME Kana mode
VK_HANGUL = 0x15                # IME Hangul mode
VK_JUNJA = 0x17                 # IME Junja mode
VK_FINAL = 0x18                 # IME final mode
VK_HANJA = 0x19                 # IME Hanja mode
VK_KANJI = 0x19                 # IME Kanji mode
VK_ESCAPE = 0x1B                # ESC key
VK_CONVERT = 0x1C               # IME convert
VK_NONCONVERT = 0x1D            # IME nonconvert
VK_ACCEPT = 0x1E                # IME accept
VK_MODECHANGE = 0x1F            # IME mode change request
VK_SPACE = 0x20                 # SPACEBAR
VK_PRIOR = 0x21                 # PAGE UP key
VK_NEXT = 0x22                  # PAGE DOWN key
VK_END = 0x23                   # END key
VK_HOME = 0x24                  # HOME key
VK_LEFT = 0x25                  # LEFT ARROW key
VK_UP = 0x26                    # UP ARROW key
VK_RIGHT = 0x27                 # RIGHT ARROW key
VK_DOWN = 0x28                  # DOWN ARROW key
VK_SELECT = 0x29                # SELECT key
VK_PRINT = 0x2A                 # PRINT key
VK_EXECUTE = 0x2B               # EXECUTE key
VK_SNAPSHOT = 0x2C              # PRINT SCREEN key
VK_INSERT = 0x2D                # INS key
VK_DELETE = 0x2E                # DEL key
VK_HELP = 0x2F                  # HELP key
VK_LWIN = 0x5B                  # Left Windows key (Natural keyboard)
VK_RWIN = 0x5C                  # Right Windows key (Natural keyboard)
VK_APPS = 0x5D                  # Applications key (Natural keyboard)
VK_SLEEP = 0x5F                 # Computer Sleep key
VK_NUMPAD0 = 0x60               # Numeric keypad 0 key
VK_NUMPAD1 = 0x61               # Numeric keypad 1 key
VK_NUMPAD2 = 0x62               # Numeric keypad 2 key
VK_NUMPAD3 = 0x63               # Numeric keypad 3 key
VK_NUMPAD4 = 0x64               # Numeric keypad 4 key
VK_NUMPAD5 = 0x65               # Numeric keypad 5 key
VK_NUMPAD6 = 0x66               # Numeric keypad 6 key
VK_NUMPAD7 = 0x67               # Numeric keypad 7 key
VK_NUMPAD8 = 0x68               # Numeric keypad 8 key
VK_NUMPAD9 = 0x69               # Numeric keypad 9 key
VK_MULTIPLY = 0x6A              # Multiply key
VK_ADD = 0x6B                   # Add key
VK_SEPARATOR = 0x6C             # Separator key
VK_SUBTRACT = 0x6D              # Subtract key
VK_DECIMAL = 0x6E               # Decimal key
VK_DIVIDE = 0x6F                # Divide key
VK_F1 = 0x70                    # F1 key
VK_F2 = 0x71                    # F2 key
VK_F3 = 0x72                    # F3 key
VK_F4 = 0x73                    # F4 key
VK_F5 = 0x74                    # F5 key
VK_F6 = 0x75                    # F6 key
VK_F7 = 0x76                    # F7 key
VK_F8 = 0x77                    # F8 key
VK_F9 = 0x78                    # F9 key
VK_F10 = 0x79                   # F10 key
VK_F11 = 0x7A                   # F11 key
VK_F12 = 0x7B                   # F12 key
VK_F13 = 0x7C                   # F13 key
VK_F14 = 0x7D                   # F14 key
VK_F15 = 0x7E                   # F15 key
VK_F16 = 0x7F                   # F16 key
VK_F17 = 0x80                   # F17 key
VK_F18 = 0x81                   # F18 key
VK_F19 = 0x82                   # F19 key
VK_F20 = 0x83                   # F20 key
VK_F21 = 0x84                   # F21 key
VK_F22 = 0x85                   # F22 key
VK_F23 = 0x86                   # F23 key
VK_F24 = 0x87                   # F24 key
VK_NUMLOCK = 0x90               # NUM LOCK key
VK_SCROLL = 0x91                # SCROLL LOCK key
VK_LSHIFT = 0xA0                # Left SHIFT key
VK_RSHIFT = 0xA1                # Right SHIFT key
VK_LCONTROL = 0xA2              # Left CONTROL key
VK_RCONTROL = 0xA3              # Right CONTROL key
VK_LMENU = 0xA4                 # Left MENU key
VK_RMENU = 0xA5                 # Right MENU key
VK_BROWSER_BACK = 0xA6          # Browser Back key
VK_BROWSER_FORWARD = 0xA7       # Browser Forward key
VK_BROWSER_REFRESH = 0xA8       # Browser Refresh key
VK_BROWSER_STOP = 0xA9          # Browser Stop key
VK_BROWSER_SEARCH = 0xAA        # Browser Search key
VK_BROWSER_FAVORITES = 0xAB     # Browser Favorites key
VK_BROWSER_HOME = 0xAC          # Browser Start and Home key
VK_VOLUME_MUTE = 0xAD           # Volume Mute key
VK_VOLUME_DOWN = 0xAE           # Volume Down key
VK_VOLUME_UP = 0xAF             # Volume Up key
VK_MEDIA_NEXT_TRACK = 0xB0      # Next Track key
VK_MEDIA_PREV_TRACK = 0xB1      # Previous Track key
VK_MEDIA_STOP = 0xB2            # Stop Media key
VK_MEDIA_PLAY_PAUSE = 0xB3      # Play/Pause Media key
VK_LAUNCH_MAIL = 0xB4           # Start Mail key
VK_LAUNCH_MEDIA_SELECT = 0xB5   # Select Media key
VK_LAUNCH_APP1 = 0xB6           # Start Application 1 key
VK_LAUNCH_APP2 = 0xB7           # Start Application 2 key
VK_OEM_1 = 0xBA                 # Used for miscellaneous characters; it can vary by keyboard.
                                # For the US standard keyboard, the ';:' key
VK_OEM_PLUS = 0xBB              # For any country/region, the '+' key
VK_OEM_COMMA = 0xBC             # For any country/region, the ',' key
VK_OEM_MINUS = 0xBD             # For any country/region, the '-' key
VK_OEM_PERIOD = 0xBE            # For any country/region, the '.' key
VK_OEM_2 = 0xBF                 # Used for miscellaneous characters; it can vary by keyboard.
                                # For the US standard keyboard, the '/?' key
VK_OEM_3 = 0xC0                 # Used for miscellaneous characters; it can vary by keyboard.
                                # For the US standard keyboard, the '`~' key
VK_OEM_4 = 0xDB                 # Used for miscellaneous characters; it can vary by keyboard.
                                # For the US standard keyboard, the '[{' key
VK_OEM_5 = 0xDC                 # Used for miscellaneous characters; it can vary by keyboard.
                                # For the US standard keyboard, the '\|' key
VK_OEM_6 = 0xDD                 # Used for miscellaneous characters; it can vary by keyboard.
                                # For the US standard keyboard, the ']}' key
VK_OEM_7 = 0xDE                 # Used for miscellaneous characters; it can vary by keyboard.
                                # For the US standard keyboard, the 'single-quote/double-quote' key
VK_OEM_8 = 0xDF                 # Used for miscellaneous characters; it can vary by keyboard.
VK_OEM_102 = 0xE2               # Either the angle bracket key or the backslash key on the RT 102-key keyboard
VK_PROCESSKEY = 0xE5            # IME PROCESS key
VK_PACKET = 0xE7                # Used to pass Unicode characters as if they were keystrokes. The VK_PACKET key is the low word of a 32-bit Virtual Key value used for non-keyboard input methods. For more information, see Remark in KEYBDINPUT, SendInput, WM_KEYDOWN, and WM_KEYUP
VK_ATTN = 0xF6                  # Attn key
VK_CRSEL = 0xF7                 # CrSel key
VK_EXSEL = 0xF8                 # ExSel key
VK_EREOF = 0xF9                 # Erase EOF key
VK_PLAY = 0xFA                  # Play key
VK_ZOOM = 0xFB                  # Zoom key
VK_PA1 = 0xFD                   # PA1 key
VK_OEM_CLEAR = 0xFE             # Clear key

WHEEL_DELTA = 120
XBUTTON1 = 0x0001
XBUTTON2 = 0x0002



import math
from enum import Enum

# Tuple structures
# midi_event = (type, id, velocity)

OCTAVE_SIZE = 12
C1_SEMITONE = 2 * OCTAVE_SIZE
KEYS = ["C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"]
DYNAMICS = [
    (16, "ppp"),
    (33, "pp"),
    (49, "p"),
    (64, "mp"),
    (80, "mf"),
    (96, "f"),
    (112, "ff"),
    (127, "fff")
]
PEDALS = [
    (64, "Sustain")
]


class EventType(Enum):
    NONE = 0
    NOTE = 144
    PEDAL = 176


def get_note_tone(midi_event):
    raw_semitone = midi_event[1]
    offset_semitone = raw_semitone - C1_SEMITONE  # center notes to be around C1
    octave = math.floor(offset_semitone / OCTAVE_SIZE)
    note_index = offset_semitone % OCTAVE_SIZE
    return KEYS[note_index] + str(octave)


def get_note_dynamic(midi_event):
    loudness = midi_event[2]
    if loudness == 0:
        return "OFF"
    else:
        for (value, name) in DYNAMICS:
            if loudness <= value:
                return name
        return "N/A"


def get_note(midi_event):
    tone = get_note_tone(midi_event)
    dynamic = get_note_dynamic(midi_event)
    return tone, dynamic


def get_pedal(midi_event):
    pedal_id = midi_event[1]
    pedal_vel = midi_event[2]

    pedal_name = "NA"
    is_down = False

    for (id_, name) in PEDALS:
        if pedal_id == id_:
            pedal_name = name
            break

    if pedal_vel == 127:
        is_down = True
    return pedal_name, is_down


def get_event_type(midi_event):
    try:
        return EventType(midi_event[0])
    except (TypeError, ValueError):
        return EventType.NONE



import sys
import os
import keys
import emulator
import pygame
import json
from pygame import  midi


def print_device_info():
    for i in range(pygame.midi.get_count()):
        info = pygame.midi.get_device_info(i)
        (interface, name, input_, output, open) = info

        if input_:
            io_status = "input"
        if output:
            io_status = "output"

        print("%2i: %s: %s (%s)" %
              (i, interface, name, io_status))


def parse_key(key_str):
    upper_str = key_str.upper()
    upper_concat1 = "KEY_" + upper_str
    upper_concat2 = "VK_" + upper_str

    for attempt in [upper_str, upper_concat1, upper_concat2]:
        result = getattr(keys, attempt, None)
        if result is not None and isinstance(result, int):
            return result

    return int(key_str)


def load_config(config_path):
    file = open(config_path, "r")
    config = json.load(file)

    device_id = config["device_id"]
    raw_bindings = config["bindings"]

    bindings = []
    for raw_binding in raw_bindings:
        note = raw_binding["note"]
        raw_key = raw_binding["key"]
        key = parse_key(raw_key)
        bindings.append(emulator.Binding(note, key))

    return device_id, bindings


def main():
    emulator.init_pygame()

    args = sys.argv[1:]

    if "-devices" in args:
        print_device_info()
        sys.exit()

    verbose = "-verbose" in args

    if len(args) < 1:
        print("A config file must be supplied")
        sys.exit(-1)
    config_path = args[-1]

    (device_id, mappings) = load_config(config_path)
    em = emulator.Emulator(device_id, mappings)
    em.verbose = verbose

    em.run()


if __name__ == "__main__":
    main()



MOUSEEVENTF_ABSOLUTE = 0x8000
MOUSEEVENTF_HWHEEL = 0x01000
MOUSEEVENTF_MOVE = 0x0001
MOUSEEVENTF_MOVE_NOCOALESCE = 0x2000
MOUSEEVENTF_LEFTDOWN = 0x0002
MOUSEEVENTF_LEFTUP = 0x0004
MOUSEEVENTF_RIGHTDOWN = 0x0008
MOUSEEVENTF_RIGHTUP = 0x0010
MOUSEEVENTF_MIDDLEDOWN = 0x0020
MOUSEEVENTF_MIDDLEUP = 0x0040
MOUSEEVENTF_VIRTUALDESK = 0x4000
MOUSEEVENTF_WHEEL = 0x0800
MOUSEEVENTF_XDOWN = 0x0080
MOUSEEVENTF_XUP = 0x0100


KEYEVENTF_EXTENDEDKEY = 0x0001
KEYEVENTF_KEYUP = 0x0002
KEYEVENTF_SCANCODE = 0x0008
KEYEVENTF_UNICODE = 0x0004


import pygame
import os
import midi
import key_handling
import key_events
from pygame import midi as pygame_midi


class Binding:
    def __init__(self, note_name, keyboard_key):
        self.note = note_name
        self.key = keyboard_key
        self.is_pressed = False

    def press(self):
        self.is_pressed = True
        key_input = key_handling.new_key_input(self.key)
        key_handling.send_input(key_input)

    def release(self):
        self.is_pressed = False
        key_input = key_handling.new_key_input(self.key, key_events.KEYEVENTF_KEYUP)
        key_handling.send_input(key_input)


class Emulator:
    def __init__(self, midi_device_id, bindings):
        self.midi_device_id = midi_device_id
        self.midi_device = pygame.midi.Input(midi_device_id)
        self.bindings = bindings
        self.verbose = False

    def release_all(self):
        for binding in self.bindings:
            binding.un_press()

    def run(self):
        running = True
        while running:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    running = False
            if self.midi_device.poll():
                midi_events = self.midi_device.read(10)
                for (midi_event, timestamp) in midi_events:
                    event_type = midi.get_event_type(midi_event)
                    if event_type == midi.EventType.PEDAL:
                        self.handle_pedal(midi_event)
                    elif event_type == midi.EventType.NOTE:
                        self.handle_note(midi_event)
        pygame.quit()

    def handle_pedal(self, midi_event):
        (pedal, is_pressed) = midi.get_pedal(midi_event)
        if is_pressed:
            self.release_all()

    def handle_note(self, midi_event):
        (note, dynamic) = midi.get_note(midi_event)
        for binding in self.bindings:
            if binding.note == note:
                if dynamic == "OFF":
                    binding.release()
                else:
                    binding.press()


def init_pygame():
    os.environ["SDL_VIDEODRIVER"] = "dummy"
    pygame.init()
    pygame.midi.init()




import ctypes

LONG = ctypes.c_long
DWORD = ctypes.c_ulong
ULONG_PTR = ctypes.POINTER(DWORD)
WORD = ctypes.c_ushort

INPUT_MOUSE = 0
INPUT_KEYBOARD = 1
INPUT_HARDWARE = 2


class MouseInput(ctypes.Structure):
    _fields_ = (('dx', LONG),
                ('dy', LONG),
                ('mouseData', DWORD),
                ('dwFlags', DWORD),
                ('time', DWORD),
                ('dwExtraInfo', ULONG_PTR))

    def __init__(self, x, y, flags, data):
        super(MouseInput, self).__init__(x, y, data, flags, 0, None)


class KeyboardInput(ctypes.Structure):
    _fields_ = (('wVk', WORD),
                ('wScan', WORD),
                ('dwFlags', DWORD),
                ('time', DWORD),
                ('dwExtraInfo', ULONG_PTR))

    def __init__(self, code, flags):
        super(KeyboardInput, self).__init__(code, code, flags, 0, None)


class HardwareInput(ctypes.Structure):
    _fields_ = (('uMsg', DWORD),
                ('wParamL', WORD),
                ('wParamH', WORD))

    def __init__(self, message, parameter):
        super(HardwareInput, self).__init__(message & 0xFFFFFFFF, parameter & 0xFFFF, parameter >> 16 & 0xFFFF)


class _InputUnion(ctypes.Union):
    _fields_ = (('mi', MouseInput),
                ('ki', KeyboardInput),
                ('hi', HardwareInput))


class Input(ctypes.Structure):
    _fields_ = (('type', DWORD),
                ('union', _InputUnion))

    def __init__(self, structure):
        if isinstance(structure, MouseInput):
            super(Input, self).__init__(INPUT_MOUSE, _InputUnion(mi=structure))
        elif isinstance(structure, KeyboardInput):
            super(Input, self).__init__(INPUT_KEYBOARD, _InputUnion(ki=structure))
        elif isinstance(structure, HardwareInput):
            super(Input, self).__init__(INPUT_HARDWARE, _InputUnion(hi=structure))
        else:
            raise TypeError('Cannot create Input structure!')


def send_input(*inputs):
    n_inputs = len(inputs)
    lp_input = Input * n_inputs
    p_inputs = lp_input(*inputs)
    cb_size = ctypes.c_int(ctypes.sizeof(Input))
    return ctypes.windll.user32.SendInput(n_inputs, p_inputs, cb_size)


def new_mouse_input(flags, x=0, y=0, data=0):
    return Input(MouseInput(flags, x, y, data))


def new_key_input(code, flags=0):
    return Input(KeyboardInput(code, flags))


def new_hardware_input(message, parameter=0):
    return Input(HardwareInput(message, parameter))



from setuptools import setup


def readme():
    with open('README.md') as f:
        return f.read()


setup(name='pyanoboard',
      version='1.0.0',
      description='Simple piano to keyboard adaptor. Works with any MIDI enabled device.',
      long_description=readme(),
      long_description_content_type='text/markdown',
      url='https://github.com/QFSW/Pyanoboard',
      author='QFSW',
      author_email='support@qfsw.co.uk',
      license='MIT',
      packages=['pyanoboard'],
      install_requires=[
            'pygame',
      ],
      entry_points={
            "console_scripts": [
                  "pyanoboard=pyanoboard.__main__:main",
            ]
      })
